@@ -3,6 +3,7 @@ package Module::Install::GetProgramLocations;
 use strict;
 use Config;
 use Cwd;
+use Carp;
 use File::Spec;
 use Sort::Versions;
 use Exporter();
@@ -12,10 +13,11 @@ use vars qw( @ISA $VERSION @EXPORT );
 use Module::Install::Base;
 @ISA = qw( Module::Install::Base Exporter );
 
-@EXPORT = qw( &Get_GNU_GPP_Version &Get_GNU_Grep_Version
-              &Get_GNU_Make_Version &Get_Bzip2_Version );
+@EXPORT = qw( &Get_GNU_Version
+              &Get_Bzip2_Version
+            );
 
-$VERSION = sprintf "%d.%02d%02d", q/0.20.0/ =~ /(\d+)/g;
+$VERSION = sprintf "%d.%02d%02d", q/0.30.0/ =~ /(\d+)/g;
 
 # ---------------------------------------------------------------------------
 
@@ -24,6 +26,12 @@ sub Get_Program_Locations
   my $self = shift;
   my %info = %{ shift @_ };
 
+  foreach my $program (keys %info)
+  {
+    croak "argname is required for $program"
+      unless defined $info{$program}{'argname'};
+  }
+
   # Module::Install says it requires perl 5.004
   $self->requires( perl => '5.004' );
   $self->include_deps('Config',0);
@@ -31,18 +39,57 @@ sub Get_Program_Locations
   $self->include_deps('Sort::Versions',0);
   $self->include_deps('Cwd',0);
 
-  # By default the programs have no paths
-  my %programs = map { $_ => undef } keys %info;
+  my %user_specified_program_paths =
+    $self->_Get_User_Specified_Program_Locations(\%info);
 
-  my ($programs_ref,$program_specified_on_command_line) =
-    $self->_Get_ARGV_Program_Locations(\%programs,\%info);
-  %programs = %$programs_ref;
+  if (keys %user_specified_program_paths)
+  {
+    return $self->_Get_ARGV_Program_Locations(\%info,
+      \%user_specified_program_paths);
+  }
+  else
+  {
+    return $self->_Prompt_User_For_Program_Locations(\%info);
+  }
+}
 
-  return %programs if $program_specified_on_command_line;
+# ---------------------------------------------------------------------------
 
-  %programs = $self->_Prompt_User_For_Program_Locations(\%programs,\%info);
+sub _Get_User_Specified_Program_Locations
+{
+  my $self = shift;
+  my %info = %{ shift @_ };
 
-  return %programs;
+  my %user_specified_program_paths;
+  my @remaining_args;
+
+  # Look for user-provided paths in @ARGV
+  foreach my $arg (@ARGV)
+  {
+    my ($var,$value) = $arg =~ /^(.*?)=(.*)$/;
+
+    push(@remaining_args, $arg), next unless defined $var;
+
+    $value = undef if $value eq '';
+
+    my $is_a_program_arg = 0;
+
+    foreach my $program (keys %info)
+    {
+      if ($var eq $info{$program}{'argname'})
+      {
+        $user_specified_program_paths{$program} = $value;
+        $is_a_program_arg = 1;
+        last;
+      }
+    }
+
+    push @remaining_args, $arg unless $is_a_program_arg;
+  }
+
+  @ARGV = @remaining_args;
+
+  return %user_specified_program_paths;
 }
 
 # ---------------------------------------------------------------------------
@@ -50,43 +97,48 @@ sub Get_Program_Locations
 sub _Get_ARGV_Program_Locations
 {
   my $self = shift;
-  my %programs = %{ shift @_ };
   my %info = %{ shift @_ };
+  my %program_locations = %{ shift @_ };
 
-  my $program_specified_on_command_line = 0;
-  my @remaining_args;
+  my %program_info;
 
-  # Look for user-provided paths in @ARGV
-  foreach my $arg (@ARGV)
+  foreach my $program_name (sort keys %info)
   {
-    my ($var,$value) = $arg =~ /^(.*?)=(.*)$/;
-    $value = undef if $value eq '';
+    $program_info{$program_name} = 
+      { 'path' => undef, 'type' => undef, 'version' => undef };
 
-    if (!defined $var)
+    next if exists $program_locations{$program_name} &&
+      $program_locations{$program_name} eq '';
+
+    $program_locations{$program_name} = $info{$program_name}{'default'}
+      unless exists $program_locations{$program_name};
+
+    my $full_path = $self->_Make_Absolute($program_locations{$program_name});
+    if (!defined $self->can_run($full_path))
     {
-      push @remaining_args, $arg;
+      warn "\"$full_path\" does not appear to be a valid executable\n";
+      warn "Using anyway\n";
+
+      $program_info{$program_name} =
+        { path => $full_path, type => undef, version => undef };
     }
     else
     {
-      my $is_a_program_arg = 0;
-
-      foreach my $program (keys %info)
+      my ($is_valid,$type,$version) = 
+        $self->_Program_Version_Is_Valid($program_name,$full_path,\%info);
+      
+      unless($is_valid)
       {
-        if ($var eq $info{$program}{'argname'})
-        {
-          $programs{$program} = $value;
-          $program_specified_on_command_line = 1;
-          $is_a_program_arg = 1;
-        }
+        warn "\"$full_path\" is not a correct version\n";
+        warn "Using anyway\n";
       }
 
-      push @remaining_args, $arg unless $is_a_program_arg;
+      $program_info{$program_name} =
+        { path => $full_path, type => $type, version => $version };
     }
   }
 
-  @ARGV = @remaining_args;
-
-  return (\%programs,$program_specified_on_command_line);
+  return %program_info;
 }
 
 # ---------------------------------------------------------------------------
@@ -94,16 +146,15 @@ sub _Get_ARGV_Program_Locations
 sub _Prompt_User_For_Program_Locations
 {
   my $self = shift;
-  my %programs = %{ shift @_ };
   my %info = %{ shift @_ };
 
-  my @path = split /$Config{path_sep}/, $ENV{PATH};
-
   print "Enter the full path, or \"none\" for none.\n";
 
   my $last_choice = '';
 
-  ASK: foreach my $program_name (sort keys %programs)
+  my %program_info;
+
+  ASK: foreach my $program_name (sort keys %info)
   {
     my ($name,$full_path);
 
@@ -119,32 +170,34 @@ sub _Prompt_User_For_Program_Locations
 
     $full_path = 'none' if !defined $full_path || $name eq '';
 
-    my $allowed_versions = '';
-    if (exists $info{$program_name}{'versions'})
+    my $allowed_types = '';
+    if (exists $info{$program_name}{'types'})
     {
-      foreach my $type (keys %{ $info{$program_name}{'versions'} } )
+      foreach my $type (keys %{ $info{$program_name}{'types'} } )
       {
-        $allowed_versions .= ", $type";
+        $allowed_types .= ", $type";
       }
 
-      $allowed_versions =~ s/^, //;
-      $allowed_versions =~ s/(.*), /$1, or /;
-      $allowed_versions = " ($allowed_versions";
-      $allowed_versions .=
-        scalar(keys %{ $info{$program_name}{'versions'} }) > 1 ?
-        " versions)" : " version)";
+      $allowed_types =~ s/^, //;
+      $allowed_types =~ s/(.*), /$1, or /;
+      $allowed_types = " ($allowed_types";
+      $allowed_types .= scalar(keys %{ $info{$program_name}{'types'} }) > 1 ?
+        " types)" : " type)";
     }
 
     my $choice = $self->prompt(
-      "Where can I find your \"$program_name\" executable?$allowed_versions" => $full_path);
+      "Where can I find your \"$program_name\" executable?" .
+      "$allowed_types" => $full_path);
 
-    $programs{$program_name} = undef, next if $choice eq 'none';
+    $program_info{$program_name} =
+      { path => undef, type => undef, version => undef }, next
+      if $choice eq 'none';
 
     $choice = $self->_Make_Absolute($choice);
 
     if (!defined $self->can_run($choice))
     {
-      print "\"$choice\" does not appear to be a valid executable\n";
+      warn "\"$choice\" does not appear to be a valid executable\n";
 
       if ($last_choice ne $choice)
       {
@@ -152,25 +205,32 @@ sub _Prompt_User_For_Program_Locations
         redo ASK;
       }
 
-      print "Using anyway\n";
+      warn "Using anyway\n";
     }
-    elsif(!$self->_Program_Version_Is_Valid($program_name,$choice,\%info))
+    else
     {
-      print "\"$choice\" is not a correct version\n";
-
-      if ($last_choice ne $choice)
+      my ($is_valid,$type,$version) = 
+        $self->_Program_Version_Is_Valid($program_name,$choice,\%info);
+      
+      if(!$is_valid)
       {
-        $last_choice = $choice;
-        redo ASK;
+        warn "\"$choice\" is not a correct version\n";
+
+        if ($last_choice ne $choice)
+        {
+          $last_choice = $choice;
+          redo ASK;
+        }
+
+        warn "Using anyway\n";
       }
 
-      print "Using anyway\n";
+      $program_info{$program_name} =
+        { path => $choice, type => undef, version => $version };
     }
-
-    $programs{$program_name} = $choice;
   }
 
-  return %programs;
+  return %program_info;
 }
 
 # ---------------------------------------------------------------------------
@@ -182,38 +242,38 @@ sub _Program_Version_Is_Valid
   my $program = shift;
   my %info = %{ shift @_ };
 
-  if (exists $info{$program_name}{'versions'})
+  if (exists $info{$program_name}{'types'})
   {
     my $program_version;
 
-    VERSION: foreach my $version (keys %{$info{$program_name}{'versions'}})
+    TYPE: foreach my $version (keys %{$info{$program_name}{'types'}})
     {
       $program_version = 
-        &{$info{$program_name}{'versions'}{$version}{'fetch'}}($program);
+        &{$info{$program_name}{'types'}{$version}{'fetch'}}($program);
 
-      next VERSION unless defined $program_version;
+      next TYPE unless defined $program_version;
 
       if ($self->Version_Matches_Range($program_version,
-        $info{$program_name}{'versions'}{$version}{'numbers'}))
+        $info{$program_name}{'types'}{$version}{'numbers'}))
       {
-        return 1;
+        return (1,$version,$program_version);
       }
     }
 
     my $program_version_string = '<UNKNOWN>';
     $program_version_string = $program_version if defined $program_version;
-    print "\"$program\" version $program_version_string is not valid for any of the following:\n";
+    warn "\"$program\" version $program_version_string is not valid for any of the following:\n";
 
-    foreach my $version (keys %{$info{$program_name}{'versions'}})
+    foreach my $version (keys %{$info{$program_name}{'types'}})
     {
-      print "  $version => " .
-        $info{$program_name}{'versions'}{$version}{'numbers'} . "\n";
+      warn "  $version => " .
+        $info{$program_name}{'types'}{$version}{'numbers'} . "\n";
     }
 
-    return 0;
+    return (0,undef,undef);
   }
 
-  return 1;
+  return (1,undef,undef);
 }
 
 # ---------------------------------------------------------------------------
@@ -255,6 +315,7 @@ sub Version_Matches_Range
 
 # ---------------------------------------------------------------------------
 
+# Returns the original if the full path can't be found
 sub _Make_Absolute
 {
   my $self = shift;
@@ -274,8 +335,10 @@ sub _Make_Absolute
       last if defined $self->can_run($path_to_choice);
     }
 
-    print "WARNING: Avoid security risks by converting to absolute paths\n";
-    print "\"$program\" is currently in your path at \"$path_to_choice\"\n";
+    return $program unless -e $path_to_choice;
+
+    warn "WARNING: Avoiding security risks by converting to absolute paths\n";
+    warn "\"$program\" is currently in your path at \"$path_to_choice\"\n";
 
     return $path_to_choice;
   }
@@ -285,6 +348,7 @@ sub _Make_Absolute
 
 sub Get_GNU_Version
 {
+  my $self = shift;
   my $program = shift;
 
   die "Missing GNU program to get version for" unless defined $program;
@@ -314,36 +378,6 @@ sub Get_GNU_Version
 
 # ---------------------------------------------------------------------------
 
-sub Get_GNU_GPP_Version
-{
-  my $self = shift;
-  my $program = shift;
-
-  return Get_GNU_Version($program);
-}
-
-# ---------------------------------------------------------------------------
-
-sub Get_GNU_Grep_Version
-{
-  my $self = shift;
-  my $program = shift;
-
-  return Get_GNU_Version($program);
-}
-
-# ---------------------------------------------------------------------------
-
-sub Get_GNU_Make_Version
-{
-  my $self = shift;
-  my $program = shift;
-
-  return Get_GNU_Version($program);
-}
-
-# ---------------------------------------------------------------------------
-
 sub Get_Bzip2_Version
 {
   my $self = shift;
@@ -382,62 +416,91 @@ A simple example:
     # A no-path default and a command line variable
     'gzip'     => { default => 'gzip', argname => 'GZIP' },
   );
-  my %locations = Get_Program_Locations(\%info);
+
+  my %location_info = Get_Program_Locations(\%info);
+
+  print "grep path is " . $location_info{'grep'}{'path'} . "\n";
 
 A complex example showing all the bells and whistles:
 
   use inc::Module::Install;
-  # So Module::Install::GetProgramLocations can be found
-  use lib 'inc';
   ...
+  # User-defined get version program
+  sub Get_Solaris_Grep_Version
+  {
+    my $program = shift;
+
+    my $result = `strings $program | $program SUNW_OST_OSCMD`;
+
+    return undef unless $result =~ /SUNW_OST_OSCMD/;
+
+    # Solaris grep isn't versioned, so we'll just return 0 for it
+    return 0;
+  }
+
   my %info = (
     # Either the GNU or the Solaris version
     'grep'     => { default => 'grep', argname => 'GREP',
-                    versions => {
+                    type => {
                       # Any GNU version higher than 2.1
-                      'GNU' =>     { fetch => \&Get_GNU_Grep_Version,
+                      'GNU' =>     { fetch => \&Get_GNU_Version,
                                      numbers => '[2.1,)', },
-                      # Any solaris version higher than 1.0, except 1.1
+                      # Any solaris version
                       'Solaris' => { fetch => \&Get_Solaris_Grep_Version,
-                                     numbers => '[1.0,1.1) (1.1,]', },
+                                     numbers => '[0,)', },
                     },
                   },
   );
-  my %locations = Get_Program_Locations(\%info);
+
+  my %location_info = Get_Program_Locations(\%info);
+
+  print "grep path is " . $location_info{'grep'}{'path'} . "\n";
+  print "grep type is " . $location_info{'grep'}{'type'} . "\n";
+  print "grep version is " . $location_info{'grep'}{'version'} . "\n";
 
 
 =head1 DESCRIPTION
 
 If you are installing a module that calls external programs, it's best to make
 sure that those programs are installed and working correctly. This
-Module::Install extension helps with that process. It allows the user to
-interactively specify either full path to a program, or an unqualified program
-name. In the latter case, the module will search the user's path to find the
-first matching program. 
+Module::Install extension helps with that process. Given a specification of
+the required programs, it attempts to find a working version on the system
+based on the Perl configuration and the user's path.  The extension then
+returns a hash mapping the program names to a hash containing the absolute
+path to the program, the type, and the version number. (It's best to use the
+absolute path in order to avoid security problems.)
 
-This extension will then perform validation on the program.  It makes sure
-that the program can be run, and will optionally check the version for
-correctness if the user provides that information. If the program can't be run
-or is the wrong version, an error message is displayed and the user is
-prompted again.
+The program specification allows the user to specify a default program, a
+command-line name for the program to be set, and multiple types of satisfying
+implementations of the program. For the types, the user can specify a function
+to extract the version, and a version range to check the version.
 
-If the user enters the same information twice, then the information is used
-regardless of any problems. The user can also enter "none" to indicate that
-the program is not available.
+The extension defaults to interactive mode, where it asks the user to specify
+the paths to the programs. If the user specifies a relative path, the
+extension converts this to an absolute path. The user can specify
+"none", in which case the hash values will be undefined. Similarly, if the , or if the type or version cannot be determined, then the
+hash values will be undefined.
 
-The extension returns a hash mapping the program names to their absolute
-paths.  (It's best to use the absolute path in order to avoid security
-problems.) If the user entered "none", then the hash value will be undefined.
+The extension also supports a noninteractive mode, where the programs are
+provided on the command line. For example, "perl Makefile.PL
+PROGRAM=<program>" is used on the command line to indicate the desired
+program. The path is converted to an absolute path.  "<program>" can be empty
+to indicate that it is not available.
 
-The user can avoid the interactive prompts by specifying one or more paths to
-the programs on the command line call to "perl Makefile.PL". Note that no
-validation of the programs is done in this case. 
+This extension will perform validation on the program, whether or not it was
+specified interactively. It makes sure that the program can be run, and will
+optionally check the version for correctness if the user provides that
+information. If the program can't be run or is the wrong version, an error
+message is displayed. In interactive mode, the user is prompted again.  If the
+user enters the same information twice, then the information is used
+regardless of any problems. In noninteractive mode, the program is used
+anyway.
 
 =head1 METHODS
 
 =over 4
 
-=item %paths = Get_Program_Locations(\%info)
+=item %program_information = Get_Program_Locations(\%info)
 
 This function takes as input a hash with information for the programs to be
 found. The keys are the program names (and can actually be anything). The
@@ -457,7 +520,7 @@ The command line variable name. For example, if you want the user to be able
 to set the path to bzip2, you might set this to "BZIP2" so that the user can
 run "perl Makefile.PL BZIP2=/usr/bin/bzip2".
 
-=item versions
+=item types
 
 A hash mapping a descriptive version name to a hash containing a mapping for
 two keys:
@@ -482,11 +545,34 @@ they compare.
 
 =back
 
+Each of your fetch routines should only succeed for one kind of
+implementation, returning undef when they fail.
+
 =back
 
 The return value for Get_Program_Locations is a hash whose keys are the same
-as those of %info, and whose values are either an absolute path for the
-program location, or undef (for no program location).
+as those of %info, and whose values are hashes having three values:
+
+=over 2
+
+=item path
+
+The absolute path to the program. This value is undef if the user chose no
+program.
+
+=item type
+
+The name of the type, if there are multiple possible types. This name
+corresponds to the name originally given in the "types" hash. This value is
+undef if no program was chosen by the user, or if there was no types hash
+value.
+
+=item version
+
+The version number of the program.
+
+=back
+
 
 =item $boolean = Version_Matches_Range($program_version, $range);
 
@@ -505,10 +591,12 @@ Feel free to submit new version functions for programs that you use.
 
 =over 4
 
-=item $version = Get_GNU_Grep_Version($path_to_program)
+=item $version = Get_GNU_Version($path_to_program)
 
-Gets the version of GNU grep. Returns undef if the word "GNU" does not appear
-in the version information
+Gets the version of a general GNU program. Returns undef if the application
+does not appear to be a GNU application. This function relies on certain
+conventions that the Free Software Foundation uses for printing the version of
+GNU applications. It may not work for all programs.
 
 =item $version = Get_Bzip2_Version($path_to_program)
 
